#include <ace/Thread_Manager.h>
#include <ace/VRT_Wrapper.h>
#include <ace/Virtual_Real_Time.h>
#include <ace/Message_Block.h>


#if defined (ACE_HAS_VIRTUAL_TIME)

void
ACE_VRT_Synchronizer::synchronize (void )
{
  // do nothing in the default case
}

/*------------------------------------------------------------------------
  The concept of Virtual Time is that it is the time spend within a single
  process and can be used to test code designed for dedicated processors
  on a multi-user system.  Virtual time is the sum of the real time that
  the process started and an estimate of the time the process comsumed 
  since it started.  The latter has two components, the actual time that
  the process used + an estimate for the time the process was idle waiting
  for an external event.

  There are only a limited number of OS functions that need to be 
  simulated in virtual time.  The basic mechanism is to wrap the real
  call in a simulation of Virtual Time.  But since we will still need
  real timeouts on the processor and they may not correspond to the 
  amount of time actually used; we actually call the function in a loop
  converging on the Virtual Time specified.

  We need to keep track of the threads in a process so that we can know
  when our process was completely blocked on waiting for something.  Since
  the OS can't tell us how many thread are in the process, we use the
  ACE_Thread_Manager.  For this to work, a restriction is placed on the
  process that only one manager exists, it is the singleton, and all
  threads, including the main thread, are registered with it.  Note that
  this prevents the common pattern of using ACE_Thread_Manager::wait();
  in the unregistered main routine, since it can never return.
------------------------------------------------------------------------*/

ACE_Virtual_Real_Time *ACE_Virtual_Real_Time::vrt_;

ACE_Virtual_Real_Time::ACE_Virtual_Real_Time(void)
  :real_start_time_(0,0),
   suspended_time_ (0,0),
   idle_time_(0,0),
   scale_factor_ (1.0),
   cpu_count_ (1),
   start_clock_time_ (0),
   test_group_ (99),
   test_thr_mgr_ (0),
   owner_ (ACE_Thread::self()),
   secondary_ (ACE_Thread::self()),

   vrt_enabled_ (0),
   synch_delay_ (0,10000),
   delete_sync_ (0),
   synchronizer_ (0)
{
  struct tms buf;
  ::times(&buf);
  start_clock_time_ = buf.tms_utime + buf.tms_stime;
  real_start_time_ = ACE_VRT_Wrapper::getrealtimeofday() 
    - current_tick_time(); //current_process_time();
}

ACE_Virtual_Real_Time::~ACE_Virtual_Real_Time (void)
{
  if (delete_sync_)
    delete synchronizer_;
}

ACE_Virtual_Real_Time*
ACE_Virtual_Real_Time::instance(void)
{
  if (ACE_Virtual_Real_Time::vrt_ == 0)
    {
      // Perform Double-Checked Locking Optimization.
      ACE_MT (ACE_GUARD_RETURN (ACE_Recursive_Thread_Mutex, ace_mon,
                                *ACE_Static_Object_Lock::instance (), 0));

      if (ACE_Virtual_Real_Time::vrt_ == 0)
        {
          ACE_NEW_RETURN (ACE_Virtual_Real_Time::vrt_,
                          ACE_Virtual_Real_Time,
                          0);
        }
    }
  return ACE_Virtual_Real_Time::vrt_;
}


int
ACE_Virtual_Real_Time::svc (void)
{
  ACE_Message_Block *mb;
  int mtype = ACE_Message_Block::MB_STOP;
  vrt_enabled_ = 0;
  this->set_owner();

  while (1) {
    if (vrt_enabled_ == 0) {
      //      ACE_DEBUG ((LM_DEBUG,"(%t) VRT suspended, waiting for message\n"));
      if (this->getq(mb) != -1) { // wait forever
        mtype = mb->msg_type();
        delete mb;
      }
      if (mtype == ACE_Message_Block::MB_HANGUP)
        break; // exit the loop
      else if (mtype == ACE_Message_Block::MB_STOP)
        continue; // suspend called again, just carry on.
      else  // assume a start message
        vrt_enabled_ = 1;
    }
    //    ACE_DEBUG ((LM_DEBUG,"(%t) VRT running\n"));

    ACE_Time_Value deadline = ACE_OS::gettimeofday() + synch_delay_;
    this->resume_test_threads();
    if (this->getq(mb,&deadline) != -1) {
      //      ACE_DEBUG ((LM_DEBUG,"(%t) VRT interrupted with message or timeout\n"));
      mtype = mb->msg_type();
      delete mb;
    }
    this->suspend_test_threads();
    if (mtype == ACE_Message_Block::MB_HANGUP)
      break;
    else if (mtype == ACE_Message_Block::MB_STOP)
      vrt_enabled_ = 0;

    if (vrt_enabled_)
      this->synchronizer()->synchronize(); // <--- TEST SPECIFIC HOOK

  }
  //  ACE_DEBUG ((LM_DEBUG,"(%t) VRT test terminated\n"));

  this->release_test_threads();

  return 0;
}

void 
ACE_Virtual_Real_Time::run_test(void)
{
  ACE_Message_Block *mb;
  ACE_NEW (mb,
           ACE_Message_Block (0,(int)ACE_Message_Block::MB_START)
           );
  this->putq(mb);
}

void
ACE_Virtual_Real_Time::pause_test(void)
{
  ACE_Message_Block *mb;
  ACE_NEW(mb,
          ACE_Message_Block (0,(int)ACE_Message_Block::MB_STOP)
          );
  this->putq(mb);
}

void
ACE_Virtual_Real_Time::stop_test(void)
{
  ACE_Message_Block *mb;
  ACE_NEW(mb,
          ACE_Message_Block (0,(int)ACE_Message_Block::MB_HANGUP)
          );
  this->putq(mb);
}

#endif /* ACE_HAS_VIRTUAL_TIME */

#if !defined (__ACE_INLINE__)
#include "ace/Virtual_Real_Time.inl"
#endif /* __ACE_INLINE__ */









